using System;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles; 

namespace USB_HID
{

	#region Custom exception
	/// <summary>
	/// Generic HID device exception
	/// </summary>
    public class HIDDeviceException : ApplicationException
    {
        public HIDDeviceException(string strMessage) : base(strMessage) { }

        public static HIDDeviceException GenerateWithWinError(string strMessage)
        {
            return new HIDDeviceException(string.Format("Msg:{0} WinEr:{1:X8}", strMessage, Marshal.GetLastWin32Error()));
        }
    }
	#endregion

    public class HIDBtn
    {
        private bool _State_Current;
        public bool State_Current
        {
            get
            {
                return _State_Current;
            }

            set
            {
               
                State_Prev = _State_Current;
               _State_Current = value;
               State_Changed = (_State_Current != State_Prev);
            }
        }

        public bool State_Prev;
        public bool State_Changed;
    }

    public class HIDAxe
    {
        private short _State_Current;
        public short State_Current
        {
            get
            {
                return _State_Current;
            }

            set
            {
                
                State_Prev = _State_Current;
                _State_Current = value;
                State_Changed = (_State_Current != State_Prev);
            }
        }

        public short State_Prev;
        public bool State_Changed;
    }


    public abstract class HIDStates
    {
        public int NbBtn;
        public int  NbAxe;

        public HIDBtn[] Btn;
        public HIDAxe[] Axe;
        public short ProfilNum;
        public bool IsRBC9Present;

        public void HitInit()
        {
            int i;
            Btn = new HIDBtn[NbBtn];
            Axe = new HIDAxe[NbAxe];

            for (i = 0; i < NbBtn; Btn[i++] = new HIDBtn());
            for (i = 0; i < NbAxe; Axe[i++] = new HIDAxe()) ;

            //IsRBC9Present = false;
            IsRBC9Present = true;
        }

        public void HitInit(int vNbBtn, int vNbAxe)
        {
            NbBtn = vNbBtn;
            NbAxe = vNbAxe;
            HitInit();
        }
       
    }

	/// <summary>
	/// Abstract HID device : Derive your new device controller class from this
	/// </summary>
    public abstract class HIDDevice : Win32Usb, IDisposable
    {
		#region Privates variables
		/// <summary>Filestream we can use to read/write from</summary>
        private FileStream m_oFile;
		/// <summary>Length of input report : device gives us this</summary>
		private int m_nInputReportLength;
		/// <summary>Length if output report : device gives us this</summary>
		private int m_nOutputReportLength;
		/// <summary>Handle to the device</summary>
		private IntPtr m_hHandle;

		#endregion

        #region Public variables
        public int nVid;
        public int nPid;

        #endregion

        #region IDisposable Members
        /// <summary>
		/// Dispose method
		/// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
		/// <summary>
		/// Disposer called by both dispose and finalise
		/// </summary>
		/// <param name="bDisposing">True if disposing</param>
        protected virtual void Dispose(bool bDisposing)
        {
            if (bDisposing)	// if we are disposing, need to close the managed resources
            {
                if (m_oFile != null)
                {
                    m_oFile.Close();
                    m_oFile = null;
                }
            }
            if (m_hHandle != IntPtr.Zero)	// Dispose and finalize, get rid of unmanaged resources
            {
                CloseHandle(m_hHandle);
            }
        }
        #endregion

		#region Privates/protected
		/// <summary>
		/// Initialises the device
		/// </summary>
		/// <param name="strPath">Path to the device</param>
		private void Initialise(string strPath)
		{
			// Create the file from the device path
            m_hHandle = CreateFile(strPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, IntPtr.Zero);
			if ( m_hHandle != InvalidHandleValue )	// if the open worked...
			{
				IntPtr lpData;
				if (HidD_GetPreparsedData(m_hHandle, out lpData))	// get windows to read the device data into an internal buffer
				{
                    try
                    {
                        HidCaps oCaps;
                        HidP_GetCaps(lpData, out oCaps);	// extract the device capabilities from the internal buffer
                        m_nInputReportLength = oCaps.InputReportByteLength;	// get the input...
                        m_nOutputReportLength = oCaps.OutputReportByteLength;	// ... and output report lengths
                        m_oFile = new FileStream(m_hHandle, FileAccess.Read | FileAccess.Write, true, m_nInputReportLength, true);	// wrap the file handle in a .Net file stream
                        //m_oFile = new FileStream(new SafeFileHandle(m_hHandle, false), FileAccess.Read | FileAccess.Write, m_nInputReportLength, true);	// wrap the file handle in a .Net file stream
                        BeginAsyncRead();	// kick off the first asynchronous read
                    }
                    catch
                    {
                        Console.WriteLine("Erreur 55");
                    }
					finally
					{
                       
						HidD_FreePreparsedData(ref lpData);	// before we quit the funtion, we must free the internal buffer reserved in GetPreparsedData
					}
				}
				else	// GetPreparsedData failed? Chuck an exception
				{
					throw HIDDeviceException.GenerateWithWinError("GetPreparsedData failed");
				}
			}
			else	// File open failed? Chuck an exception
			{
				m_hHandle = IntPtr.Zero;
				throw HIDDeviceException.GenerateWithWinError("Failed to create device file");
			}
		}
		/// <summary>
		/// Kicks off an asynchronous read which completes when data is read or when the device
		/// is disconnected. Uses a callback.
		/// </summary>
        private void BeginAsyncRead()
        {
            byte[] arrInputReport = new byte[m_nInputReportLength];
			// put the buff we used to receive the stuff as the async state then we can get at it when the read completes
     
            m_oFile.BeginRead(arrInputReport, 0, m_nInputReportLength, new AsyncCallback(ReadCompleted), arrInputReport);
    

        }
		/// <summary>
		/// Callback for above. Care with this as it will be called on the background thread from the async read
		/// </summary>
		/// <param name="iResult">Async result parameter</param>
        protected void ReadCompleted(IAsyncResult iResult)
        {
            byte[] arrBuff = (byte[])iResult.AsyncState;	// retrieve the read buffer
            try
            {
                
                m_oFile.EndRead(iResult);	// call end read : this throws any exceptions that happened during the read
 
                try
                {
					InputReport oInRep = CreateInputReport();	// Create the input report for the device
					oInRep.SetData(arrBuff);	// and set the data portion - this processes the data received into a more easily understood format depending upon the report type
                    HandleDataReceived(oInRep);	// pass the new input report on to the higher level handler
                }
                finally
                {
                  //  Console.WriteLine("Erreur 2");
       
                    BeginAsyncRead();	// when all that is done, kick off another read for the next report

                }                
            }
            catch(IOException )	// if we got an IO exception, the device was removed
            {
                Console.WriteLine("Erreur 1");
                HandleDeviceRemoved();
                Dispose();
            }
        }
		/// <summary>
		/// Write an output report to the device.
		/// </summary>
		/// <param name="oOutRep">Output report to write</param>
        protected void Write(OutputReport oOutRep)
        {
           	try
			{
				m_oFile.Write(oOutRep.Buffer, 0, oOutRep.BufferLength);
                unsafe
                {
                 
                    bool Success = HidD_SetOutputReport(m_hHandle, ref oOutRep.Buffer[0], oOutRep.BufferLength);
                }
                
			}
                
			catch(IOException )
           // catch (Exception ex)
			{
                Console.WriteLine("Erreur 3");
				// The device was removed!
				throw new HIDDeviceException("Device was removed");
			}
        }
		/// <summary>
		/// virtual handler for any action to be taken when data is received. Override to use.
		/// </summary>
		/// <param name="oInRep">The input report that was received</param>
		protected virtual void HandleDataReceived(InputReport oInRep)
		{
		}
		/// <summary>
		/// Virtual handler for any action to be taken when a device is removed. Override to use.
		/// </summary>
		protected virtual void HandleDeviceRemoved()
		{
          
		}
		/// <summary>
		/// Helper method to return the device path given a DeviceInterfaceData structure and an InfoSet handle.
		/// Used in 'FindDevice' so check that method out to see how to get an InfoSet handle and a DeviceInterfaceData.
		/// </summary>
		/// <param name="hInfoSet">Handle to the InfoSet</param>
		/// <param name="oInterface">DeviceInterfaceData structure</param>
		/// <returns>The device path or null if there was some problem</returns>
        private static string GetDevicePath(IntPtr hInfoSet, ref SP_DEVICE_INTERFACE_DATA oInterface)
   
        {
            Int32 nRequiredSize = 0;

            // Get the device interface details
            SP_DEVINFO_DATA odevData = new SP_DEVINFO_DATA();

            if (!SetupDiGetDeviceInterfaceDetail(hInfoSet, oInterface, IntPtr.Zero, 0, ref nRequiredSize, odevData))
			{
                IntPtr buffer = Marshal.AllocHGlobal(nRequiredSize);
                SP_DEVICE_INTERFACE_DETAIL_DATA oDetail = new SP_DEVICE_INTERFACE_DETAIL_DATA();
                oDetail.cbSize = Marshal.SizeOf(typeof(SP_DEVICE_INTERFACE_DETAIL_DATA));
                oDetail.cbSize = 8;
                Marshal.StructureToPtr(oDetail, buffer, false);
                if (!SetupDiGetDeviceInterfaceDetail(hInfoSet, oInterface, buffer, nRequiredSize, ref nRequiredSize, odevData))
                {
                    System.ComponentModel.Win32Exception t = new System.ComponentModel.Win32Exception(Marshal.GetLastWin32Error());
                    MessageBox(new IntPtr(0), t.Message, "Error", 0);
                    Marshal.FreeHGlobal(buffer);
                    return null;
                }

                IntPtr pDevicePath = (IntPtr)((int)buffer + Marshal.SizeOf(typeof(int)));
                string devicePath = Marshal.PtrToStringAuto(pDevicePath);
                Marshal.FreeHGlobal(buffer);

                return devicePath;

			}
			return null;
		}

        private static short GetDeviceUsage(string strPath)
        {
            IntPtr m_hHandle;
            short vUsage = -1;

            // Create the file from the device path
            m_hHandle = CreateFile(strPath, 0, FILE_SHARE_READ | FILE_SHARE_WRITE , IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
            if (m_hHandle != InvalidHandleValue)	// if the open worked...
            {
                IntPtr lpData;
                if (HidD_GetPreparsedData(m_hHandle, out lpData))	// get windows to read the device data into an internal buffer
                {
                    try
                    {
                        HidCaps oCaps;
                        HidP_GetCaps(lpData, out oCaps);	// extract the device capabilities from the internal buffer
                        vUsage = oCaps.Usage;
                    }
                    catch
                    {
                        Console.WriteLine("Erreur 55");
                    }
                    finally
                    {

                        HidD_FreePreparsedData(ref lpData);	// before we quit the funtion, we must free the internal buffer reserved in GetPreparsedData
                    }
                }
                else	// GetPreparsedData failed? Chuck an exception
                {
                    throw HIDDeviceException.GenerateWithWinError("GetPreparsedData failed");
                }

                CloseHandle(m_hHandle);
                m_hHandle = IntPtr.Zero;
            }
            else	// File open failed? Chuck an exception
            {
                m_hHandle = IntPtr.Zero;
                throw HIDDeviceException.GenerateWithWinError("Failed to create device file");
            }

            return vUsage;
        }

        static bool StateChange(IntPtr hDevInfo, int SelectedItem, int NewState)
        {

            SP_PROPCHANGE_PARAMS PropChangeParams;

            SP_DEVINFO_DATA2 DeviceInfoData;

            PropChangeParams = new SP_PROPCHANGE_PARAMS();

            PropChangeParams.Init();

            DeviceInfoData = new SP_DEVINFO_DATA2();

            PropChangeParams.ClassInstallHeader.cbSize = Marshal.SizeOf(PropChangeParams.ClassInstallHeader);

            DeviceInfoData.cbSize = Marshal.SizeOf(DeviceInfoData);

            if (!SetupDiEnumDeviceInfo(hDevInfo, SelectedItem, ref DeviceInfoData))
                return false;

            PropChangeParams.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE;
            PropChangeParams.Scope = DICS_FLAG_GLOBAL;
            PropChangeParams.StateChange = NewState;

            if (!SetupDiSetClassInstallParams(hDevInfo,
                ref DeviceInfoData,
                ref PropChangeParams.ClassInstallHeader,
                Marshal.SizeOf(PropChangeParams)))
                return false;

            if (!SetupDiCallClassInstaller(Win32Usb.DIF_PROPERTYCHANGE, hDevInfo, ref DeviceInfoData))
                return false;

            return true;
        }
  
/*  exemple restrat driver ??
        BOOLEAN CSetupDIMgr::RestartDevice(void) 
{ 
    CString sHardwareID; 
    if(HardwareID(sHardwareID)) 
    { 
        do 
        { 
            SP_PROPCHANGE_PARAMS params; 

            // for future compatibility; this will zero out the entire struct, rather 
            // than just the fields which exist now 
            memset(&params, 0, sizeof(SP_PROPCHANGE_PARAMS)); 

            // initialize the SP_CLASSINSTALL_HEADER struct at the beginning of the 
            // SP_PROPCHANGE_PARAMS struct, so that SetupDiSetClassInstallParams will 
            // work 
            params.ClassInstallHeader.cbSize = sizeof(SP_CLASSINSTALL_HEADER); 

            params.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE; 
          
             // initialize SP_PROPCHANGE_PARAMS such that the device will be stopped. 
            params.StateChange = DICS_PROPCHANGE; 
            params.Scope       = DICS_FLAG_CONFIGSPECIFIC; 
            params.HwProfile   = 0; // current profile 

            if(!SetupDiSetClassInstallParams(m_devInfo, &m_devInfoData, 
                (PSP_CLASSINSTALL_HEADER) &params, 
                sizeof(SP_PROPCHANGE_PARAMS))) 
            { 
                TRACE("in RestartDevice(): couldn't set the install parameters!"); 
                TRACE(" error: %u\n", GetLastError()); 
                return FALSE; 
            } 

            if(!SetupDiCallClassInstaller(DIF_PROPERTYCHANGE, m_devInfo, &m_devInfoData)) 
            { 
                  TRACE("in RestartDevice(): call to class installer (STOP) failed!"); 
                  TRACE(" error: %u\n", GetLastError() ); 
                  return FALSE; 
            } 
        } 
        while(FindHardwareID(sHardwareID, m_nDevIndex)); 
        return TRUE; 
     } 
     return FALSE; 
}*/
		#endregion

		#region Public static
        /// <summary>
		/// Finds a device given its PID and VID
		/// </summary>
		/// <param name="nVid">Vendor id for device (VID)</param>
		/// <param name="nPid">Product id for device (PID)</param>
		/// <param name="oType">Type of device class to create</param>
		/// <returns>A new device class of the given type or null</returns>
		public static HIDDevice FindDevice(int nVid, int nPid, short nUsage,Type oType)
        {
            string strPath = string.Empty;
			string strSearch = string.Format("vid_{0:x4}&pid_{1:x4}", nVid, nPid); // first, build the path search string
            Guid gHid;
            HidD_GetHidGuid(out gHid);	// next, get the GUID from Windows that it uses to represent the HID USB interface
            IntPtr hInfoSet = SetupDiGetClassDevs(ref gHid, null, IntPtr.Zero, DIGCF_DEVICEINTERFACE | DIGCF_PRESENT);	// this gets a list of all HID devices currently connected to the computer (InfoSet)
            try
            {
                SP_DEVICE_INTERFACE_DATA oInterface = new SP_DEVICE_INTERFACE_DATA();	// build up a device interface data block
            
                int nIndex = 0;
                while (SetupDiEnumDeviceInterfaces(hInfoSet, null, ref gHid, nIndex, oInterface))	// this gets the device interface information for a device at index 'nIndex' in the memory block
                {
                    string strDevicePath = GetDevicePath(hInfoSet, ref oInterface);	// get the device path (see helper method 'GetDevicePath')
                    if (strDevicePath.IndexOf(strSearch) >= 0)	// do a string search, if we find the VID/PID string then we found our device!
                    {

                        if (GetDeviceUsage(strDevicePath) == nUsage) 
                        {
                            HIDDevice oNewDevice = (HIDDevice)Activator.CreateInstance(oType);	// create an instance of the class for this device
                            oNewDevice.Initialise(strDevicePath);	// initialise it with the device path
                            oNewDevice.nVid =nVid;
                                oNewDevice.nPid = nPid;
                            SetupDiDestroyDeviceInfoList(hInfoSet);
                            return oNewDevice;	// and return it
                        }
                    }
                    nIndex++;	// if we get here, we didn't find our device. So move on to the next one.
                }
            }
            catch (HIDDeviceException)
            {
                Console.WriteLine("Erreur 4");
               // throw new HIDDeviceException("IO Error");

            }

            finally
            {
				// Before we go, we have to free up the InfoSet memory reserved by SetupDiGetClassDevs
                SetupDiDestroyDeviceInfoList(hInfoSet);
            }
            return null;	// oops, didn't find our device
        }


		#endregion

		#region Publics
		/// <summary>
		/// Accessor for output report length
		/// </summary>
		public int OutputReportLength
		{
			get
			{
				return m_nOutputReportLength;
			}
		}
		/// <summary>
		/// Accessor for input report length
		/// </summary>
		public int InputReportLength
		{
			get
			{
				return m_nInputReportLength;
			}
		}
		/// <summary>
		/// Virtual method to create an input report for this device. Override to use.
		/// </summary>
		/// <returns>A shiny new input report</returns>
		public virtual InputReport CreateInputReport()
		{
			return null;
		}
		#endregion
    }
}
